#!/usr/bin/env python3

# This file enables editable installs with `pip install --editable .`
# https://setuptools.readthedocs.io/en/latest/setuptools.html#setup-cfg-only-projects

import setuptools

setuptools.setup()
